@@ -29,20 +29,12 @@ would be appreciated.
 
 """
 
-from __future__ import (absolute_import, division, print_function)
-
 from random import randint
 import warnings
-import six
-from six.moves import filter
-try:                    # Python 3
-    from urllib.parse import (urlencode, urlparse, urlunparse, parse_qs,
-                              ParseResult)
-except ImportError:      # Python 2
-    from urllib import urlencode
-    from urlparse import urlparse, urlunparse, parse_qs, ParseResult
+from urllib.parse import (urlencode, urlparse, urlunparse, parse_qs,
+                          ParseResult)
 from .etree import etree
-from .util import openURL, testXMLValue, getXMLInteger
+from .util import clean_ows_url, testXMLValue, getXMLInteger, Authentication, openURL
 from .fgdc import Metadata
 from .iso import MD_Metadata
 from .ows import ServiceProvider, ServiceIdentification, OperationsMetadata
@@ -52,6 +44,9 @@ _OWS_NS = '{http://www.opengis.net/ows/1.1}'
 _WMTS_NS = '{http://www.opengis.net/wmts/1.0}'
 _XLINK_NS = '{http://www.w3.org/1999/xlink}'
 
+# OpenGIS Web Map Tile Service (WMTS) Implementation Standard
+# Version 1.0.0, document 07-057r7
+
 _ABSTRACT_TAG = _OWS_NS + 'Abstract'
 _IDENTIFIER_TAG = _OWS_NS + 'Identifier'
 _LOWER_CORNER_TAG = _OWS_NS + 'LowerCorner'
@@ -77,7 +72,11 @@ _MIN_TILE_ROW_TAG = _WMTS_NS + 'MinTileRow'
 _RESOURCE_URL_TAG = _WMTS_NS + 'ResourceURL'
 _SCALE_DENOMINATOR_TAG = _WMTS_NS + 'ScaleDenominator'
 _SERVICE_METADATA_URL_TAG = _WMTS_NS + 'ServiceMetadataURL'
+
+# Table 7, page 20-21, Parts of Style data structure
 _STYLE_TAG = _WMTS_NS + 'Style'
+_STYLE_LEGEND_URL = _WMTS_NS + 'LegendURL'
+
 _THEME_TAG = _WMTS_NS + 'Theme'
 _THEMES_TAG = _WMTS_NS + 'Themes'
 _TILE_HEIGHT_TAG = _WMTS_NS + 'TileHeight'
@@ -127,9 +126,9 @@ class WebMapTileService(object):
         else:
             raise KeyError("No content named %s" % name)
 
-    def __init__(self, url, version='1.0.0', xml=None, username=None,
-                 password=None, parse_remote_metadata=False,
-                 vendor_kwargs=None):
+    def __init__(self, url, version='1.0.0', xml=None, username=None, password=None,
+                 parse_remote_metadata=False, vendor_kwargs=None, headers=None, auth=None,
+                 timeout=30):
         """Initialize.
 
         Parameters
@@ -150,19 +149,28 @@ class WebMapTileService(object):
         vendor_kwargs : dict
             Optional vendor-specific parameters to be included in all
             requests.
+        auth : owslib.util.Authentication
+            Instance of Authentication class to hold username/password/cert/verify
+        timeout : int
+            number of seconds for GetTile request
 
         """
-        self.url = url
-        self.username = username
-        self.password = password
+        self.url = clean_ows_url(url)
+        if auth:
+            if username:
+                auth.username = username
+            if password:
+                auth.password = password
         self.version = version
         self.vendor_kwargs = vendor_kwargs
         self._capabilities = None
+        self.headers = headers
+        self.auth = auth or Authentication(username, password)
+        self.timeout = timeout or 30
 
         # Authentication handled by Reader
-        reader = WMTSCapabilitiesReader(self.version, url=self.url,
-                                        un=self.username, pw=self.password)
-
+        reader = WMTSCapabilitiesReader(
+            self.version, url=self.url, headers=self.headers, auth=self.auth)
         if xml:  # read from stored xml
             self._capabilities = reader.readString(xml)
         else:  # read from server
@@ -180,12 +188,13 @@ class WebMapTileService(object):
         self._buildMetadata(parse_remote_metadata)
 
     def _getcapproperty(self):
+        # TODO: deprecated function. See ticket #453.
         if not self._capabilities:
             reader = WMTSCapabilitiesReader(
-                self.version, url=self.url, un=self.username, pw=self.password
-                )
-            xml = reader.read(self.url, self.vendor_kwargs)
-            self._capabilities = ServiceMetadata(xml)
+                self.version, url=self.url, headers=self.headers, auth=self.auth)
+            # xml = reader.read(self.url, self.vendor_kwargs)
+            # self._capabilities = ServiceMetadata(xml)
+            self._capabilities = reader.read(self.url, self.vendor_kwargs)
         return self._capabilities
 
     def _buildMetadata(self, parse_remote_metadata=False):
@@ -195,7 +204,8 @@ class WebMapTileService(object):
 
         # serviceIdentification metadata
         serviceident = self._capabilities.find(_SERVICE_IDENTIFICATION_TAG)
-        self.identification = ServiceIdentification(serviceident)
+        if serviceident is not None:
+            self.identification = ServiceIdentification(serviceident)
 
         # serviceProvider metadata
         serviceprov = self._capabilities.find(_SERVICE_PROVIDER_TAG)
@@ -218,7 +228,7 @@ class WebMapTileService(object):
         def gather_layers(parent_elem, parent_metadata):
             for index, elem in enumerate(parent_elem.findall(_LAYER_TAG)):
                 cm = ContentMetadata(
-                    elem, parent=parent_metadata, index=index+1,
+                    elem, parent=parent_metadata, index=index + 1,
                     parse_remote_metadata=parse_remote_metadata)
                 if cm.id:
                     if cm.id in self.contents:
@@ -287,19 +297,6 @@ class WebMapTileService(object):
             Column index of tile to request.
         **kwargs : extra arguments
             anything else e.g. vendor specific parameters
-
-        Example
-        -------
-            >>> url = 'http://map1c.vis.earthdata.nasa.gov/wmts-geo/wmts.cgi'
-            >>> wmts = WebMapTileService(url)
-            >>> wmts.buildTileRequest(layer='VIIRS_CityLights_2012',
-            ...                       tilematrixset='EPSG4326_500m',
-            ...                       tilematrix='6',
-            ...                       row=4, column=4)
-            'SERVICE=WMTS&REQUEST=GetTile&VERSION=1.0.0&\
-LAYER=VIIRS_CityLights_2012&STYLE=default&TILEMATRIXSET=EPSG4326_500m&\
-TILEMATRIX=6&TILEROW=4&TILECOL=4&FORMAT=image%2Fjpeg'
-
         """
 
         if (layer is None):
@@ -314,9 +311,9 @@ TILEMATRIX=6&TILEROW=4&TILECOL=4&FORMAT=image%2Fjpeg'
             msg = 'tilematrix (zoom level) is mandatory (cannot be None)'
             raise ValueError(msg)
         if row is None:
-                raise ValueError("row is mandatory (cannot be None)")
+            raise ValueError("row is mandatory (cannot be None)")
         if column is None:
-                raise ValueError("column is mandatory (cannot be None)")
+            raise ValueError("column is mandatory (cannot be None)")
 
         request = list()
         request.append(('SERVICE', 'WMTS'))
@@ -330,7 +327,7 @@ TILEMATRIX=6&TILEROW=4&TILECOL=4&FORMAT=image%2Fjpeg'
         request.append(('TILECOL', str(column)))
         request.append(('FORMAT', format))
 
-        for key, value in six.iteritems(kwargs):
+        for key, value in list(kwargs.items()):
             request.append((key, value))
 
         data = urlencode(request, True)
@@ -340,6 +337,21 @@ TILEMATRIX=6&TILEROW=4&TILECOL=4&FORMAT=image%2Fjpeg'
                           tilematrixset=None, tilematrix=None, row=None,
                           column=None, **kwargs):
 
+        # check the validity of the parameters and set reasonable defaults
+        if layer is None:
+            raise ValueError("layer is mandatory (cannot be None)")
+        if style is None:
+            style = list(self[layer].styles.keys())[0]
+        if tilematrixset is None:
+            tilematrixset = sorted(self[layer].tilematrixsetlinks.keys())[0]
+        if tilematrix is None:
+            msg = 'tilematrix (zoom level) is mandatory (cannot be None)'
+            raise ValueError(msg)
+        if row is None:
+            raise ValueError("row is mandatory (cannot be None)")
+        if column is None:
+            raise ValueError("column is mandatory (cannot be None)")
+
         tileresourceurls = []
         for resourceURL in self[layer].resourceURLs:
             if resourceURL['resourceType'] == 'tile':
@@ -349,13 +361,11 @@ TILEMATRIX=6&TILEROW=4&TILECOL=4&FORMAT=image%2Fjpeg'
             # choose random ResourceURL if more than one available
             resindex = randint(0, numres - 1)
             resurl = tileresourceurls[resindex]['template']
-            if tilematrixset:
-                resurl = resurl.replace('{TileMatrixSet}', tilematrixset)
+            resurl = resurl.replace('{TileMatrixSet}', tilematrixset)
             resurl = resurl.replace('{TileMatrix}', tilematrix)
-            resurl = resurl.replace('{TileRow}', row)
-            resurl = resurl.replace('{TileCol}', column)
-            if style:
-                resurl = resurl.replace('{Style}', style)
+            resurl = resurl.replace('{TileRow}', str(row))
+            resurl = resurl.replace('{TileCol}', str(column))
+            resurl = resurl.replace('{Style}', style)
             return resurl
 
         return None
@@ -422,19 +432,6 @@ TILEMATRIX=6&TILEROW=4&TILECOL=4&FORMAT=image%2Fjpeg'
             Column index of tile to request.
         **kwargs : extra arguments
             anything else e.g. vendor specific parameters
-
-        Example
-        -------
-            >>> url = 'http://map1c.vis.earthdata.nasa.gov/wmts-geo/wmts.cgi'
-            >>> wmts = WebMapTileService(url)
-            >>> img = wmts.gettile(layer='VIIRS_CityLights_2012',\
-                                   tilematrixset='EPSG4326_500m',\
-                                   tilematrix='6',\
-                                   row=4, column=4)
-            >>> out = open('tile.jpg', 'wb')
-            >>> bytes_written = out.write(img.read())
-            >>> out.close()
-
         """
         vendor_kwargs = self.vendor_kwargs or {}
         vendor_kwargs.update(kwargs)
@@ -444,7 +441,7 @@ TILEMATRIX=6&TILEROW=4&TILECOL=4&FORMAT=image%2Fjpeg'
             resurl = self.buildTileResource(
                 layer, style, format, tilematrixset, tilematrix,
                 row, column, **vendor_kwargs)
-            u = openURL(resurl, username=self.username, password=self.password)
+            u = openURL(resurl, headers=self.headers, auth=self.auth, timeout=self.timeout)
             return u
 
         # KVP implemetation
@@ -470,14 +467,13 @@ TILEMATRIX=6&TILEROW=4&TILECOL=4&FORMAT=image%2Fjpeg'
                     base_url = get_verbs[0].get('url')
             except StopIteration:
                 pass
-        u = openURL(base_url, data, username=self.username,
-                    password=self.password)
+        u = openURL(base_url, data, headers=self.headers, auth=self.auth, timeout=self.timeout)
 
         # check for service exceptions, and return
         if u.info()['Content-Type'] == 'application/vnd.ogc.se_xml':
             se_xml = u.read()
             se_tree = etree.fromstring(se_xml)
-            err_message = six.text_type(se_tree.find('ServiceException').text)
+            err_message = str(se_tree.find('ServiceException').text)
             raise ServiceException(err_message.strip(), se_xml)
         return u
 
@@ -669,8 +665,7 @@ class ContentMetadata:
 
     Implements IContentMetadata.
     """
-    def __init__(self, elem, parent=None, index=0,
-                 parse_remote_metadata=False):
+    def __init__(self, elem, parent=None, index=0, parse_remote_metadata=False):
         if elem.tag != _LAYER_TAG:
             raise ValueError('%s should be a Layer' % (elem,))
 
@@ -702,8 +697,7 @@ class ContentMetadata:
         # probably be shared code
 
         self._tilematrixsets = [f.text.strip() for f in
-                                elem.findall(_TILE_MATRIX_SET_LINK_TAG + '/' +
-                                             _TILE_MATRIX_SET_TAG)]
+                                elem.findall(_TILE_MATRIX_SET_LINK_TAG + '/' + _TILE_MATRIX_SET_TAG)]
 
         link_elements = elem.findall(_TILE_MATRIX_SET_LINK_TAG)
         tile_matrix_set_links = TileMatrixSetLink.from_elements(link_elements)
@@ -729,18 +723,39 @@ class ContentMetadata:
             style = {}
             isdefaulttext = s.attrib.get('isDefault')
             style['isDefault'] = (isdefaulttext == "true")
-            identifier = s.find(_IDENTIFIER_TAG)
+            identifier = s.find(_IDENTIFIER_TAG)  # one and mandatory
             if identifier is None:
                 raise ValueError('%s missing identifier' % (s,))
+
             title = s.find(_TITLE_TAG)
             if title is not None:
-                style['title'] = title.text
+                style['title'] = testXMLValue(title)
+
+            abstract = s.find(_ABSTRACT_TAG)
+            if abstract is not None:
+                style['abstract'] = testXMLValue(abstract)
+
+            legendURL = s.find(_STYLE_LEGEND_URL)
+            if legendURL is not None:
+                style['legend'] = legendURL.attrib[_HREF_TAG]
+                if 'width' in list(legendURL.attrib.keys()):
+                    style['width'] = legendURL.attrib.get('width')
+                if 'height' in list(legendURL.attrib.keys()):
+                    style['height'] = legendURL.attrib.get('height')
+                if 'format' in list(legendURL.attrib.keys()):
+                    style['format'] = legendURL.attrib.get('format')
+
+            keywords = [f.text for f in s.findall(
+                        _KEYWORDS_TAG + '/' + _KEYWORD_TAG)]
+            if keywords:  # keywords is a list []
+                style['keywords'] = keywords
+
             self.styles[identifier.text] = style
 
         self.formats = [f.text for f in elem.findall(_FORMAT_TAG)]
 
         self.keywords = [f.text for f in elem.findall(
-                         _KEYWORDS_TAG+'/'+_KEYWORD_TAG)]
+                         _KEYWORDS_TAG + '/' + _KEYWORD_TAG)]
         self.infoformats = [f.text for f in elem.findall(_INFO_FORMAT_TAG)]
 
         self.layers = []
@@ -765,13 +780,18 @@ class WMTSCapabilitiesReader:
     """Read and parse capabilities document into a lxml.etree infoset
     """
 
-    def __init__(self, version='1.0.0', url=None, un=None, pw=None):
+    def __init__(self, version='1.0.0', url=None, un=None, pw=None, headers=None, auth=None):
         """Initialize"""
         self.version = version
         self._infoset = None
         self.url = url
-        self.username = un
-        self.password = pw
+        if auth:
+            if un:
+                auth.username = un
+            if pw:
+                auth.password = pw
+        self.auth = auth or Authentication(un, pw)
+        self.headers = headers
 
     def capabilities_url(self, service_url, vendor_kwargs=None):
         """Return a capabilities url
@@ -806,8 +826,7 @@ class WMTSCapabilitiesReader:
 
         # now split it up again to use the generic openURL function...
         spliturl = getcaprequest.split('?')
-        u = openURL(spliturl[0], spliturl[1], method='Get',
-                    username=self.username, password=self.password)
+        u = openURL(spliturl[0], spliturl[1], method='Get', headers=self.headers, auth=self.auth)
         return etree.fromstring(u.read())
 
     def readString(self, st):
